// Ethernet (Wiznet W5500) specific routines for "MIDI Merger"  ***************
//
// The MIDI Merger
// merges up to 4 conventional MIDI Inputs into a common MIDI message stream
// to a PC via Ethernet (Multicast or standard UDP) or via USB MIDI Class.
// Vice versa a stream of MIDI messages from a PC is distributed to
// up to 4 conventional MIDI Outputs with channel filter per output.
//
// The design is based on the STM32 Nucleo-L476 module plus a specific PCB.
//
// Author and (c)2017-18 Wolfgang Schemmert. The copyright claim and
// the GPL-3 Licence is restricted to this file and the depending files
// "main.c", "main.h", "usb-otg.c", "midi.c"  and "userinterface.c".
// Other depending material is published and copyrighted
// by STMicroelectronics, Atollic AB and EMBITZ.org
//
// Contact: <www.midi-and-more.de>
// A detailled construction and operation manual is provided at this website
//
// Note: at present state of art, only the DEBUG object code is working.
// The RELEASE version compiled with EmBitz 1.11 does not work. Reason unkown.
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************


#include "stm32l4xx.h"
#include "globaldefines.h"

//main
extern struct PARAM{
uint32_t Vid;
uint32_t Pid;
uint32_t MacHi;
uint32_t MacLo;
uint32_t Gateway;
uint32_t LocalIP;
uint32_t Subnet;
uint32_t UdpDestIP;
uint32_t UdpDestPort;
uint32_t UdpLocalPort;
uint32_t MidiIP;
uint32_t McastDestPort;
uint32_t McastLocalPort;
uint32_t LanMode;
uint32_t Midi13Mask;
uint32_t Midi45Mask;
uint32_t EthUsbMask;
uint32_t TcpPort;
uint32_t DisconTime;
uint32_t ComBaud;
uint32_t MidiBaud;
uint32_t Padding;
} gComParam;

extern uint8_t gOpMode;       //0=Setup, 1=UDP/COM, 2=TCP/COM, 0x1x=MIDI Merge
extern uint16_t gGREEN_LEDCOUNT, gTIMEOUT;
extern void wait1ms(void);
extern void wait1us(void);
extern void servSystem(void);
//USART
extern void msDelay(uint16_t) ;
//void sendSetupB(uint8_t);     //may be uncommented for debug output
//extern void sendX8(uint8_t);
//extern void sendX16(uint16_t);
//extern void sendX32(uint32_t);

//from midi.c
extern uint8_t gmBlockF6, gmBlockF8, gmBlockFE;

// ************************************

uint8_t geSPI_RXBUF[SPI_BUFLEN];
uint8_t *geSPI_RXHEAD=0,*geSPI_RXTAIL=0,*geSPI_RXBUFTOP= &(geSPI_RXBUF[SPI_BUFLEN]);
uint16_t geSPI_TXBUF[SPI_BUFLEN];
uint16_t *geSPI_TXHEAD=0,*geSPI_TXTAIL=0,*geSPI_TXBUFTOP= &(geSPI_TXBUF[SPI_BUFLEN]);
uint8_t geDummy;            //global only to avoid warning during compilation
uint8_t geWizMode;          //reflects S0_MR register of W5500
uint16_t geDiscount;

uint8_t initW5500(void);
uint8_t openUDP0(void);
uint8_t openTCP0(void);
void closeTCPConnection(void);
void clearW5500Interrupts(void);
//uint8_t addEtherQuery(uint8_t);
uint8_t checkEtherSend(void);
//uint8_t sendEtherQuery(void);
uint8_t sendEtherByte(uint8_t);
uint16_t sendEtherBuffer(uint8_t *, uint16_t);
uint8_t receiveUDP(void);
uint8_t receiveTCP(void);

void initSPI1(void);
uint16_t exchangeSpiWord(uint16_t);
uint8_t exchangeSpiByte(uint8_t);
void writeWiz_4(uint16_t, uint8_t, uint8_t);
uint8_t readWiz_4(uint16_t, uint8_t);
void writeWiz_5(uint16_t, uint8_t, uint16_t);
uint16_t readWiz_5(uint16_t, uint8_t);
void writeWiz_7(uint16_t, uint8_t, uint32_t);
uint32_t readWiz_7(uint16_t, uint8_t);
uint16_t writeWiz_Data(uint16_t, uint8_t*, uint16_t);
uint16_t readWiz_Data(uint16_t, uint16_t);
uint16_t pullSpiRxBuf(void);

/*
//W5500 Common Registers (SPI Adressing)
//--------------------------------------
#define COMMON_BLOCK    0x00
#define MR              0x0000  //Mode Register, for S/W Reset
#define GAR0            0x0001  //4 Gateway Address Register
#define SUBR0           0x0005  //4 Subnet Mask Register
#define SHAR0           0x0009  //6 Source Hardware Address Register
#define SIPR0           0x000F  //4 Source IP Address Register
#define IR              0x0015  //Interrupt Register
#define IMR             0x0016  //Interrupt Mask Register
#define SIR             0x0017  //Socket Interrupt Register
#define SIMR            0x0018  //Socket Interrupt Mask Register
#define RTR0            0x0019  //Word Initial Retry Time-value Register
#define RCR             0x001B  //Retry Count Register
#define PHYCFGR         0x002E  //PHY Configuration Register

//W5500 Socket0 Registers
//-----------------------
#define S0_REGBLOCK     0x01
#define S0_MR           0x0000  //Socket 0 Mode Register
#define S0_CR           0x0001  //Socket 0 Command Register
#define S0_IR           0x0002  //Socket 0 Interrupt Register
#define S0_SR           0x0003  //Socket 0 Status Register
#define S0_PORT0        0x0004  //Socket 0 Local Port Register
#define S0_DHAR0        0x0006  //Socket 0 Destination Hardware Address
#define S0_DIPR0        0x000C  //Socket 0 Destination IP Address
#define S0_DPORT0       0x0010  //Socket 0 Destination Port Register
#define S0_MSSR0        0x0012  //Socket 0 Maximum Segment Size
#define S0_TTL			0x0016  //Socket 0 Time to Live Register
#define S0_RXBUF_SIZE   0x001E  //Socket 0 Receive Buffer Size
#define S0_TXBUF_SIZE   0x001F  //Socket 0 Transmit Buffer Size
#define S0_TX_FSR0      0x0020  //Socket 0 TX Free Size
#define S0_TX_RD0       0x0022  //Socket 0 TX Read Pointer
#define S0_TX_WR0       0x0024  //Socket 0 TX Write Pointer
#define S0_RX_RSR0      0x0026  //Socket 0 RX Received Size
#define S0_RX_RD0       0x0028  //Socket 0 RX Read Pointer
#define S0_RX_WR0       0x002A  //Socket 0 RX Write Pointer
#define S0_IMR          0x002C  //Socket 0 Interrupt Mask Register
//#define S0_KPALVTR      0x002F  //Socket 0 Keep Alive Timer (TCP only)

#define S0_TXBLOCK      0x02
#define S0_RXBLOCK      0x03
*/

uint8_t initW5500(void)   // **********************************************
// initializes Wiz5500 registers for desired mode of operation.
// **************************************************************************
{
    GPIOC->BSRR = GPIO_BSRR_BR4;                               //set /RST low
    msDelay(5);
    GPIOC->BSRR = GPIO_BSRR_BS4;                               //set /RST high
    msDelay(100);                                               //min 50ms

    msDelay(100);
//allocate maximum socket memory (16kB)
    writeWiz_4(S0_RXBUF_SIZE,S0_REGBLOCK,16);
    writeWiz_4(S0_TXBUF_SIZE,S0_REGBLOCK,16);
//transfer MAC ID from gComParam
    writeWiz_5(SHAR0,COMMON_BLOCK,(uint16_t)(gComParam.MacHi & 0x0000FFFF));
    writeWiz_7(SHAR0+2,COMMON_BLOCK,gComParam.MacLo);
//transfer Gateway IP from gComParam
    writeWiz_7(GAR0,COMMON_BLOCK,gComParam.Gateway);
//transfer Subnet Mask from gComParam
    writeWiz_7(SUBR0,COMMON_BLOCK,gComParam.Subnet);
//transfer Local IP from gComParam
    writeWiz_7(SIPR0,COMMON_BLOCK,gComParam.LocalIP);

    writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);             //clear pending interrupts
    return 1;
}


uint8_t openUDP0(void)   // *************************************************
// with entries from gComParam
// **************************************************************************
{
uint8_t recvbyte;

//transfer UDP Local Port from gComParam
   writeWiz_5(S0_PORT0,S0_REGBLOCK,(uint16_t)gComParam.UdpLocalPort);
//transfer UDP Destination Port from gComParam
    writeWiz_5(S0_DPORT0,S0_REGBLOCK,(uint16_t)gComParam.UdpDestPort);
//transfer UDP Destination IP from gComParam
    writeWiz_7(S0_DIPR0,S0_REGBLOCK,gComParam.UdpDestIP);
//activate UDP mode
    if ((gComParam.UdpDestIP & 0xF0000000) == 0xE0000000) {
//special handling for Multicast IP
        geWizMode = 0x82;                               //internal control flag
        writeWiz_5(S0_DHAR0,S0_REGBLOCK,0x0100);
        writeWiz_7(S0_DHAR0+2,S0_REGBLOCK,0x5E000000 |
                                           (gComParam.UdpDestIP & 0x00FFFFFF));
        writeWiz_4(S0_TTL,S0_REGBLOCK,30);              //general standard value

        gTIMEOUT = 5000;
         do {                                           //wait until LINK is ok
            recvbyte = readWiz_4(PHYCFGR,COMMON_BLOCK);
            IWDG->KR = 0x0000AAAA;  //feed watchdog, timeout ca. 1-2 seconds
            if (gTIMEOUT == 0) break;
        } while ((recvbyte & 0xFD) != 0xBD);
        msDelay(100);                                  //+ little extra delay
        writeWiz_4(S0_MR,S0_REGBLOCK,0x82);             //Multicast mode
                                                //blocking Broadcast and Unicast
        msDelay(1000);                                  //+ little extra delay
    }
    else  {
//UDP Unicast or Broadcast
        geWizMode = 2;                                  //internal control flag
        writeWiz_4(S0_MR,S0_REGBLOCK,0x02);             //normal UDP mode
    }
    writeWiz_4(S0_CR,S0_REGBLOCK,0x01);                         //=CMD_OPEN
    recvbyte = 0;
    while ((recvbyte= readWiz_4(S0_SR,S0_REGBLOCK)) != 0x22 ) ;

    writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);             //clear interrupt flags
    writeWiz_4(S0_IMR,S0_REGBLOCK,0);               //no interrupts
    writeWiz_4(SIR,COMMON_BLOCK,0);
    writeWiz_4(SIMR,COMMON_BLOCK,0);

    return 1;
}


uint8_t openUDP0_MidiMerge(void)   // ***************************************
// with entries from gComParam
// **************************************************************************
{
uint8_t recvbyte;

//transfer UDP Local Port from gComParam
    writeWiz_5(S0_PORT0,S0_REGBLOCK,(uint16_t)gComParam.McastLocalPort);
//transfer UDP Destination Port from gComParam
    writeWiz_5(S0_DPORT0,S0_REGBLOCK,(uint16_t)gComParam.McastDestPort);
//transfer UDP Destination IP from gComParam
    writeWiz_7(S0_DIPR0,S0_REGBLOCK,gComParam.MidiIP);

//activate UDP mode
    if ((gComParam.MidiIP & 0xF0000000) == 0xE0000000) {
//special handling for Multicast IP
        gTIMEOUT = 5000;        //milliseconds, countdown in SysTick IR handler
         do {                                           //wait until LINK is ok
            recvbyte = readWiz_4(PHYCFGR,COMMON_BLOCK);
            IWDG->KR = 0x0000AAAA;  //feed watchdog, timeout ca. 1-2 seconds
            if (gTIMEOUT == 0) break;
        } while ((recvbyte & 0xFD) != 0xBD);

        geWizMode = 0x82;                               //internal control flag
        writeWiz_5(S0_DHAR0,S0_REGBLOCK,0x0100);
        writeWiz_7(S0_DHAR0+2,S0_REGBLOCK,0x5E000000 |
                                           (gComParam.MidiIP & 0x00FFFFFF));
        writeWiz_4(S0_TTL,S0_REGBLOCK,30);              //general standard value

        msDelay(1000);                                  //+ little extra delay
        writeWiz_4(S0_MR,S0_REGBLOCK,0x82);             //Multicast mode
    }
    else  {
//UDP Unicast or Broadcast
        geWizMode = 2;                                  //internal control flag
        writeWiz_4(S0_MR,S0_REGBLOCK,0x02);             //normal UDP mode
    }
    writeWiz_4(S0_CR,S0_REGBLOCK,0x01);                         //=CMD_OPEN
    do  recvbyte= readWiz_4(S0_SR,S0_REGBLOCK);
    while (recvbyte != 0x22 );                                  //=SOCK_UDP
    msDelay(2);

    writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);             //clear interrupt flags
    writeWiz_4(S0_IMR,S0_REGBLOCK,0);               //no interrupts
    writeWiz_4(SIR,COMMON_BLOCK,0);
    writeWiz_4(SIMR,COMMON_BLOCK,0);

    return 1;
}


uint8_t openTCP0(void)   // *************************************************
// with entries from gComParam
// **************************************************************************
{
//uint16_t recvword,n;
uint8_t recvbyte;

//init TCP mode
    writeWiz_4(S0_MR,S0_REGBLOCK,0x01);                 //TCP mode
    geWizMode = 1;                                      //internal control flag
    //transfer UDP Local Port from gComParam
    writeWiz_5(S0_PORT0,S0_REGBLOCK,(uint16_t)gComParam.TcpPort);
    writeWiz_4(S0_CR,S0_REGBLOCK,0x01);                         //=CMD_OPEN
    recvbyte = 0;
    while ((recvbyte= readWiz_4(S0_SR,S0_REGBLOCK)) != 0x13 );  //=SOCK_INIT
    wait1ms();
    writeWiz_4(S0_CR,S0_REGBLOCK,0x02);                         //=CMD_LISTEN
    while ((recvbyte= readWiz_4(S0_SR,S0_REGBLOCK)) != 0x14 );  //=SOCK_LISTEN

    clearW5500Interrupts();
    return 1;
}


void closeTCPConnection(void)   // ******************************************
// **************************************************************************
{
       writeWiz_4(S0_CR,S0_REGBLOCK,0x08);                      //=CMD_DISCON
       geDiscount = 0;
}


void clearW5500Interrupts(void)   // ***************************************
//(re-)enable S0 interrupts, only used for TCP
// **************************************************************************
{
        writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);
        writeWiz_4(S0_IMR,S0_REGBLOCK,0xFF);
        writeWiz_4(SIR,COMMON_BLOCK,0x01);           //only SOCKET0 interrupts
        writeWiz_4(SIMR,COMMON_BLOCK,0x01);
}


uint8_t sendEtherByte(uint8_t data)  // *************************************
// **************************************************************************
{
uint16_t temp, txPtr;

    if (geWizMode == 1)  {                          //internal control flag
        temp = readWiz_4(S0_SR,S0_REGBLOCK);
        if (temp != 0x17) return 0;                 //not TCP SOCK_ESTABLISHED
    }
    temp = readWiz_5(S0_TX_FSR0,S0_REGBLOCK);
    if (temp == 0) return 0;                        //no buffer available

    txPtr = readWiz_5(S0_TX_WR0,S0_REGBLOCK);
    writeWiz_4(txPtr,S0_TXBLOCK,data);
    txPtr++;
    writeWiz_5(S0_TX_WR0,S0_REGBLOCK,txPtr);        //update W5500 write pointer

    writeWiz_4(S0_CR,S0_REGBLOCK,0x20);             //=CMD_SEND
    if (geWizMode == 1) geDiscount = gComParam.DisconTime;
    writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);

    return 1;
}


uint16_t sendEtherBuffer(uint8_t *buffer, uint16_t toSend)  // **************
// demands linear buffer, not cyclic !
// **************************************************************************
{
uint16_t temp, txPtr;

    temp = readWiz_4(S0_SR,S0_REGBLOCK);
    if ((geWizMode == 1) && (temp != 0x17))  return 0; //=TCP SOCK_ESTABLISHED
    temp = readWiz_5(S0_TX_FSR0,S0_REGBLOCK);
    if (temp < toSend) return 0;                        //no buffer available

    txPtr = readWiz_5(S0_TX_WR0,S0_REGBLOCK);
    txPtr = writeWiz_Data(txPtr, buffer, toSend);
    writeWiz_5(S0_TX_WR0,S0_REGBLOCK,txPtr);        //update W5500 write pointer

    writeWiz_4(S0_CR,S0_REGBLOCK,0x20);             //=CMD_SEND
    if (geWizMode == 1) geDiscount = gComParam.DisconTime;

    return 1;
}


uint8_t receiveUDP(void)  // ************************************************
//if available, one UDP packet is transferred
//from W5500 internal receive buffer to geSPI_RXBUF
//buffer pointers are updated
// **************************************************************************
{

uint16_t rxCount, rxPtr, temp;
uint32_t udpHeader;

    rxCount = readWiz_5(S0_RX_RSR0,S0_REGBLOCK);
    if (rxCount == 0) return 1;
    if (rxCount <= 8) return 0;                     //empty UDP packet or error

    rxPtr = readWiz_5(S0_RX_RD0,S0_REGBLOCK);

    udpHeader = readWiz_7(rxPtr,S0_RXBLOCK);
    rxPtr += 4;
    udpHeader = readWiz_7(rxPtr,S0_RXBLOCK);
    rxCount -= 8;
    rxPtr += 4;
    temp =(uint16_t)(udpHeader & 0x0000FFFF); //what if temp != rxPtr ????????
    rxPtr = readWiz_Data(rxPtr,temp);
    writeWiz_5(S0_RX_RD0,S0_REGBLOCK,rxPtr);        //update W5500 read pointer

    writeWiz_4(S0_CR,S0_REGBLOCK,0x40);             //=CMD_RECV
    return 1;
}


uint8_t receiveTCP(void)  // ************************************************
// **************************************************************************
{
uint8_t temp;
uint16_t rxCount, rxPtr;

    temp = readWiz_4(S0_SR,S0_REGBLOCK);
    if (temp != 0x17)  return 0;                        //=SOCK_ESTABLISHED
    rxCount = readWiz_5(S0_RX_RSR0,S0_REGBLOCK);
    if (rxCount == 0) return 1;

    rxPtr = readWiz_5(S0_RX_RD0,S0_REGBLOCK);
    rxPtr = readWiz_Data(rxPtr,rxCount);
    writeWiz_5(S0_RX_RD0,S0_REGBLOCK,rxPtr);        //update W5500 read pointer

    writeWiz_4(S0_CR,S0_REGBLOCK,0x40);             //=CMD_RECV
    geDiscount = gComParam.DisconTime;
    return 1;
}


uint16_t pullSpiRxBuf(void) // *********************************************
// pull 1 byte asynchronously from SPI1 reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint16_t recv;

    servSystem();   //serve watchdog etc.

    if (geSPI_RXTAIL != geSPI_RXHEAD)  {
        recv= (uint16_t)((*geSPI_RXTAIL++ & 0x00FF)|0x8000);

        if (geSPI_RXTAIL == geSPI_RXBUFTOP) geSPI_RXTAIL=geSPI_RXBUF;
        if (gOpMode > 0x10){
            if ((recv == 0x80F8) && (gmBlockF8)) return 0;
            if ((recv == 0x80FE) && (gmBlockFE)) return 0;
            if ((recv == 0x80F6) && (gmBlockF6)) return 0;
        }
        GPIOC->BSRR = GPIO_BSRR_BR_2; // clear bit PC2, green LED OFF
        gGREEN_LEDCOUNT = 200; //milliseconds
        return recv;
    }
    else return 0;
}

// ##########################################################################
// low level routines

void initSPI1(void)   // ****************************************************
//SPI1 is Master
// **************************************************************************
{
     RCC->APB2ENR |=  RCC_APB2ENR_SPI1EN ;   // SPI1 clock enable

//PB14 is /INT input
//PB13 is /SCN output
    GPIOB->MODER &= ~(GPIO_MODER_MODE13 | GPIO_MODER_MODE14 );
    GPIOB->MODER |= GPIO_MODER_MODE13_0;
    GPIOB->OTYPER &= ~(GPIO_OTYPER_OT13);    //PB13 push-pull
    GPIOB->PUPDR &= ~(GPIO_PUPDR_PUPDR14);
    GPIOB->PUPDR |= GPIO_PUPDR_PUPDR14_0; //activate PB14 pull-up
//set GPIOB 3,4,5,13,14 Speed up to 25MHz ("middle speed")
    GPIOB->OSPEEDR &= 0xC3FFF03F ;  //OSPEEDRx_1 down
    GPIOB->OSPEEDR |= 0x14000540 ;  //OSPEEDRx_0 up
    GPIOB->BSRR = GPIO_BSRR_BS13;   //set /SCN high

//PC4 is /RST output
    GPIOC->MODER |= GPIO_MODER_MODE4_0;
    GPIOC->OTYPER &= ~(GPIO_OTYPER_OT4);    //PC4 push-pull
//set GPIOC 4 Speed up to 25MHz ("middle speed")
    GPIOC->OSPEEDR &= ~(GPIO_OSPEEDR_OSPEED4);  //OSPEEDR4 down
    GPIOC->OSPEEDR |= GPIO_OSPEEDR_OSPEED4_0;  //OSPEEDR4_0 up
    GPIOC->BSRR = GPIO_BSRR_BS4;    //set /RST high

//PB3 is SPI1_SCK
//PB4 is SPI1_MISO
//PB5 is SPI1-MOSI, configure as Alternate Function
    GPIOB->MODER &= ~(GPIO_MODER_MODE3 | GPIO_MODER_MODE4 | GPIO_MODER_MODE5);
    GPIOB->MODER |= (GPIO_MODER_MODE3_1|GPIO_MODER_MODE4_1|GPIO_MODER_MODE5_1);

//SPI1 is AF5
    GPIOB->AFR[0] &= 0xFF000FFF;   //configure PB3,4,5 AF function
    GPIOB->AFR[0] |= 0x00555000;

    geSPI_RXHEAD = geSPI_RXBUF;
    geSPI_RXTAIL = geSPI_RXBUF;
    geSPI_TXHEAD = geSPI_TXBUF;
    geSPI_TXTAIL = geSPI_TXBUF;

//SPI Master configuration
    SPI1->CR1 = SPI_CR1_SSM | SPI_CR1_SSI | SPI_CR1_MSTR;
                                //idle=low,sample rising edge,freq=1/2APB2 clock
    SPI1->CR2 =  SPI_CR2_DS;    //default 16bit, polling
    SPI1->CR1 |= SPI_CR1_SPE;   //enable
}

/*
//was contained in previous firmware versions and worked acceptably
//removed, because critical and somewhat obscure together with STM32

uint8_t exchangeSpiByte(uint8_t data)   // **********************************
// **************************************************************************
{
uint8_t recv;

    SPI1->CR2 = SPI_CR2_FRXTH | 0x0700 ;

    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low

    while ((SPI1->SR & SPI_SR_TXE) == 0);
    SPI1->DR = (uint16_t)data;
    while ((SPI1->SR & SPI_SR_RXNE) == 0);
    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high
//STM32 processors support 8 bit SPI transfer very badly, always 16 clock pulses
//for this reason, when data blocks with even count are transferred,
//exchangeSpiByte() must always be the last transfer command
//and /SCN must be set after 8bits are transferred
//if possible, avoid exchangeSpiByte()
    recv = SPI1->DR;
    wait1us();
//???????? special hack for fine adjustment of pulse edges ?????
//seemed necessary for a while during firmware development,
//but now it works without. Exact reason unknown.
    while (SPI1->SR & SPI_SR_RXNE) {
        geDummy = SPI1->DR;     //geDummy global to avoid compiler warning msg
    }

    SPI1->CR2 =  SPI_CR2_DS;
    return recv;
}
*/

uint16_t exchangeSpiWord(uint16_t data)   // ********************************
// **************************************************************************
{
    while ((SPI1->SR & SPI_SR_TXE) == 0);
    SPI1->DR = data;
    while ((SPI1->SR & SPI_SR_RXNE) == 0);
    return SPI1->DR;
}


void writeWiz_4(uint16_t address, uint8_t block, uint8_t data)   // *******
// write 4 byte Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t send;

    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low

    send = ((block<<11) | 0x0400 | data);
    exchangeSpiWord(address);
    exchangeSpiWord(send) ;

    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high

}


void writeWiz_5(uint16_t address, uint8_t block, uint16_t data) // **********
// write 5 byte Wiz5500 compatible cluster (1 halfword, e.g. port no.)
// **************************************************************************
{

uint16_t send;

    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low
    send = (block<<11) | 0x0400 | (data >> 8) ;
    exchangeSpiWord(address);
    exchangeSpiWord(send) ;

    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high
    send = (block<<11) | 0x0400 | (data & 0x00FF);
    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low
                                                // /SCN is high for about 1 us
    exchangeSpiWord(address+1);
    exchangeSpiWord(send) ;

    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high
}


void writeWiz_7(uint16_t address, uint8_t block, uint32_t data)  // *********
// write 7 byte Wiz5500 compatible cluster (1 word, e.g. IP address)
// **************************************************************************
{
uint16_t send;

    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low

    send = (block<<11) | 0x0400 | (uint16_t)((data >> 24) & 0x000000FF) ;
    exchangeSpiWord(address);
    exchangeSpiWord(send);

    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high
    send = (block<<11) | 0x0400 | (uint16_t)((data >> 16) & 0x000000FF);
    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low
                                                // /SCN is high for about 1 us
    exchangeSpiWord(address+1);
    exchangeSpiWord(send);

    exchangeSpiWord((uint16_t)(data & 0x0000FFFF));

    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high
}

/*
// alternative version, saves code at the cost of a small amount of runtime
void writeWiz_7(uint16_t address, uint8_t block, uint32_t data)  // *********
// write 7 byte Wiz5500 compatible cluster (1 word, e.g. IP address)
// **************************************************************************
{
    writeWiz_5(address,block,(uint16_t)((data >> 16) & 0x0000FFFF));
    writeWiz_5(address+2,block,(uint16_t)(data & 0x0000FFFF));
}
*/


uint8_t readWiz_4(uint16_t address, uint8_t block)   // *******************
// read 1 databyte Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t recv, send;

    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low

    exchangeSpiWord(address);
    send = (block<<11);
    recv = exchangeSpiWord(send) ;
    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high

    return ((uint8_t)(recv & 0x00FF));
}


uint16_t readWiz_5(uint16_t address, uint8_t block)   // ********************
// read 1 halfword Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t send, recv0, recv1;


    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low

    send = (block<<11);
    exchangeSpiWord(address);
    recv0 = exchangeSpiWord(send);

    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high
    send = (block<<11);
    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low
                                                // /SCN is high for about 0.5 us
    exchangeSpiWord(address+1);
    recv1 = exchangeSpiWord(send);

    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high

    return ((recv0 & 0x00FF) << 8) | (recv1 & 0x00FF);
}


uint32_t readWiz_7(uint16_t address, uint8_t block)   // ********************
// read 1 word Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t  send;
uint32_t recv32, recv1, recv0;

    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low

    send = (block<<11);
    exchangeSpiWord(address);
    recv0 = (uint32_t)exchangeSpiWord(send);

    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high
    send = (block<<11);
    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low
                                                // /SCN is high for about 0.5 us
    exchangeSpiWord(address+1);
    recv1 = (uint32_t)exchangeSpiWord(send);

    recv32 = (uint32_t)exchangeSpiWord(0);

    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high

    return ((recv0 <<24) | ((recv1 & 0x000000FF) <<16) | (recv32 & 0x0000FFFF));
}


/*
// alternative version, saves code at the cost of a small amount of runtime
uint32_t readWiz_7(uint16_t address, uint8_t block)   // ********************
// read 1 word Wiz5500 compatible cluster
// **************************************************************************
{
uint32_t  recv2, recv0;
    recv0 = (uint32_t)readWiz_5(address,block);
    recv2 = (uint32_t)readWiz_5(address+2,block);
    return ((recv0 << 16) | (recv2 & 0x0000FFFF));
}
*/


uint16_t writeWiz_Data(uint16_t address, uint8_t* buffer, uint16_t toSend) // *
// write (toSend+3) byte Wiz5500 compatible cluster (i.e. toSend databytes)
// **************************************************************************
{
uint16_t send;
uint8_t byte1, byte2;

    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low

    if ((toSend & 1) == 0)  {               //an even number of bytes to be sent
    send = ((S0_TXBLOCK<<11)|0x0400 |*buffer++); //linear buffer, no wrap around
    exchangeSpiWord(address);
    exchangeSpiWord(send);
    toSend--;
    address++;
    }

    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high
    send = ((S0_TXBLOCK<<11) | 0x0400 | *buffer++);

    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low
                                                // /SCN is high for about 1 us
    exchangeSpiWord(address);
    exchangeSpiWord(send);
    toSend--;
    address++;

//from here, only zero or an even number of bytes to be sent
    while (toSend)  {
        byte1 = *buffer++;
        byte2 = *buffer++;
        exchangeSpiWord(byte1<<8 | byte2);
        toSend -= 2;
        address += 2;
    }
    GPIOB->BSRR = GPIO_BSRR_BS13;                               //set /SCN high
    return address;
}


uint16_t readWiz_Data(uint16_t address,uint16_t toRead)  // *****************
// read (count) databytes from Wiz5500 Rx Buffer to geSPI_RXBUF
// **************************************************************************
{
uint16_t recvword, send;

    GPIOB->BSRR = GPIO_BSRR_BR13;                               //set /SCN low

    if ((toRead & 1) == 0)   {              //an even number of bytes to be read
        send = (S0_RXBLOCK<<11);
        exchangeSpiWord(address);
        recvword = exchangeSpiWord(send);
        *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);
        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        toRead--;
        address++;
    }

    GPIOB->BSRR = GPIO_BSRR_BS13;           //set /SCN high
    send = (S0_RXBLOCK<<11);
    send = (S0_RXBLOCK<<11);                //repeat for a little extra delay
    GPIOB->BSRR = GPIO_BSRR_BR13;           //set /SCN low
                                            // /SCN is high for about 0.5 us
    exchangeSpiWord(address);
    recvword = exchangeSpiWord(send);
    *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);
    geSPI_RXHEAD++;
    if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
    toRead--;
    address++;

//from here, only zero or an even number of bytes to be read
    while (toRead)  {
        recvword = exchangeSpiWord(0);
        *geSPI_RXHEAD = (uint8_t)((recvword >> 8) & 0x00FF);
        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);
        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        toRead -= 2;
        address += 2;
    }
    GPIOB->BSRR = GPIO_BSRR_BS13;           //set /SCN high
    return address;
}

//###########################################################################

